"""
Create experimental environment for vivado HLS 
"""

import logging
from os import chdir
from pathlib import Path
import shutil
import tempfile

from plumbum import local

from . import compedit
from . import vivado_report as vrpt

class TmpDirManager:
    def __init__(self, destruct=True):
        logger = logging.getLogger("vrs_log")
        self.destruct = destruct
        self.work_dir = Path.cwd()
        self.path = Path(tempfile.mkdtemp())
        logger.info("Creating temporary directory in {}".format(self.path))
        logger.debug("work_dir : {}".format(self.work_dir))
        logger.debug("self_destruct : {}".format(self.destruct))

    def __enter__(self):
        logging.getLogger("vrs_log").debug("Changing current dir to {}".format(self.path)) 
        chdir(self.path)

    def __exit__(self, _, __, ___):
        logger = logging.getLogger("vrs_log")
        logger.debug("leaving temporary dir and switching to {}".format(self.work_dir))
        chdir(self.work_dir)
        if self.destruct:
            logger.debug("Removing temporary directory")
            shutil.rmtree(self.path)
        pass

def generate_script(vivado_script, clock_period, standard, comp_name, defines, includes, part):
    with vivado_script.open("w") as o:
        logger = logging.getLogger("vrs_log")
        logger.debug("Starting generation of tcl script")
        o.write("open_project vivado_hls_synthesis\n")
        o.write("set_top {}\n".format(comp_name))
        includes_str = "" if includes is None else " ".join(
            ["-I{}".format(s) for s in includes]
        )
        if defines is None:
            def_str = ""
        else:
            def_str = " ".join(["-D{}={}".format(k, v) for k,v in defines.items()])
        o.write('add_files comp.cpp -cflags "-std={} {} {}"\n'.format(
                standard, 
                includes_str, 
                def_str
            )
        )
        o.write('open_solution "solution"\n')
        o.write('set_part {{{}}} -tool vivado\n'.format(part)) 
        o.write('create_clock -period {} -name default\n'.format(
                clock_period
        ))
        o.write('csynth_design\n')
        o.write('export_design -flow impl -rtl vhdl -format ip_catalog\n')

def run_script(vivado_script):
    vivado_hls = local["vivado_hls"]
    vivado_hls(str(vivado_script))


def experiment(comp_path, 
               comp_name, 
               exp_name, 
               clock_period,
               part,
               standard,
               includes = None, 
               defines = None, 
               keep_env = False,
               depth_constraint = None
            ):
    logger = logging.getLogger("vrs_log")
    logger.info("Experiment: {}".format(exp_name))
    comp = Path(comp_path).resolve()
    if not comp.exists():
        raise FileNotFounError("Error when starting experiment: component file"
                               "{} does not exist".format(comp))
    logger.info("Found component file {}".format(comp))
    with TmpDirManager(not keep_env):
        comp_copy = Path("comp.cpp")
        compedit.decorate_comp(comp, comp_copy, depth_constraint)
        logger.debug("Adding pragma decoration")

        vivado_script = Path("vivado_script.tcl")
        generate_script(vivado_script, 
                        clock_period, 
                        standard,
                        comp_name, 
                        defines,
                        includes, 
                        part
                       )
        run_script(vivado_script)
        vivado_hls_rpt = Path(
                "./vivado_hls_synthesis/solution/syn/report/{}_csynth.xml".format(comp_name)
            ).resolve()
        syn_res = vrpt.parse_syn_report(vivado_hls_rpt)
        vivado_report = Path(
                "./vivado_hls_synthesis/solution/impl/report/vhdl/{}_export.xml".format(comp_name)
            ).resolve()
        impl_res = vrpt.parse_impl_report(vivado_report)
    return {"syn" : syn_res, "impl" : impl_res}





"""
Decorate the component description to add vivado HLS pragmas
"""

from pathlib import Path

DECORATE_TAG = "//@@HINT_DECORATE@@"

def decorate_comp(filename, output_file, max_latency=None):  
    in_file = Path(filename)
    out_file = Path(output_file)
    with in_file.open("r") as f:
        with out_file.open("w") as o:
            for line in f.readlines():
                if line.strip() == DECORATE_TAG:
                    if max_latency is not None:
                        o.write(line.replace(
                            DECORATE_TAG, 
                            "#pragma HLS LATENCY MAX={}\n".format(max_latency)
                        ))
                    o.write(line.replace(
                        DECORATE_TAG, 
                        "#pragma HLS PIPELINE"
                    ))
                else:
                    o.write(line)
                    



"""
Define how to iterate over synthesis
"""

from pathlib import Path
import logging

from . import vivado_exp

class csv_handler:
    def __init__(self, filename):
        self.file = Path(filename).resolve()
        with self.file.open("w") as f:
            f.write("version, cycles, II, estimated_period, "
                    "real_period, Slices, LUT, FF, DSP, BRAM, SRL\n") 

    def handle(self, record):
        syn = record["syn"]
        impl = record["impl"]
        with self.file.open('a') as f:
            f.write("{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}\n".format(
                syn['Vivado_HLS_Version'],
                syn['Pipeline_Depth'],
                syn['II'],
                syn['estimated_period'],
                impl['timing'],
                impl['Slices'],
                impl['LUT'],
                impl['FF'],
                impl['DSP'],
                impl['BRAM'],
                impl['SRL'])
            )

def minimize_latency(
        comp_path,
        comp_name,
        exp_prefix,
        clock_period,
        part,
        standard,
        includes,
        defines,
        handlers,
        keep_envs = False
    ):
    logger = logging.getLogger("vrs_log")
    handler = logging.FileHandler("{}.log".format(exp_prefix), "w")
    handler.setLevel(logging.INFO)
    logger.addHandler(handler)
    logger.info("{} : Trying to minimize latency for top level component {} " 
                "described in {}".format(exp_prefix, comp_name, comp_path))
    
    logger.info("Starting without latency constraint")
    ret = vivado_exp.experiment(comp_path, comp_name, exp_prefix+"_base",
                                clock_period, part, standard, includes,
                                defines, keep_envs)

    constraint = float(clock_period)

    for h in handlers:
        h.handle(ret)
    timing = float(ret['impl']['timing'])
    cycles = int(ret['syn']['Pipeline_Depth'])
    logger.info("Baseline result : {} cycles at a clock period of "
                "{}ns".format(cycles, timing))
    timings = {cycles : timing}
    up = cycles
    cur_best = cycles
    cur_best_timing = timing
    low = 1
    while up > low + 1:
        mid = low +  (up - low) // 2
        logger.info("Running the experiment with a pipeline depth constraint "
                    "of {}".format(mid))
        ret = vivado_exp.experiment(comp_path, comp_name,
                                    exp_prefix+"_{}".format(mid),
                                clock_period, part, standard, includes,
                                defines, keep_envs, mid)
        for h in handlers:
            h.handle(ret)
        timing = float(ret['impl']['timing'])
        cycles = int(ret['syn']['Pipeline_Depth'])
        timings[cycles] = timing
        logger.info("Result : {} cycles of {} ns".format(cycles, timing)) 
        if timing <= constraint:
            up = mid
            cur_best = cycles
            cur_best_timing = timing
        else:
            low = mid
        logger.info("New interval : [{},{}]".format(low, up))
    logger.removeHandler(handler)
    
    best = up if (low == 1 or timings[low] > constraint) else low
    best_timing = timings[best]

    return best, best_timing



import argparse
import configparser
import logging
from pathlib import Path
from sys import stdout

from . import exp_strategy

DEFAULT_PART = "xc7k160tfbg484-1"
DEFAULT_STANDARD = "c++11"

def parse_includes(include_str, ini_path):
    preprocessed = [Path(s.strip()) for s in include_str.split(',')]
    final = []
    for p in preprocessed:
        if p.is_absolute():
            final.append(p)
        else:
            absolute = (ini_path / p).resolve()
            final.append(absolute)
    return final

def parse_defines(defines_str):
    defines_list = defines_str.split(',')
    defines = {}
    for d in defines_list:
        if '=' in d:
            d = d.split('=')
            defines[d[0].strip()] = d[1].strip()
        else:
            defines[d.strip()] = 1
    return defines

def run_metaexp(name, config, config_path):
    exp_name = name
    comp_path = config['comp_path']
    comp_name = config['top_level_comp']
    clock_period = int(config['period'])
    part = config['part'] if 'part' in config else DEFAULT_PART
    standard = config['standard'] if 'standard' in config else DEFAULT_STANDARD
    includes = parse_includes(config['includes'], config_path) if 'includes' in config else []
    defines = parse_defines(config['defines']) if 'defines' in config else {}
    keep_env = bool(config['keep_env']) if 'keep_env' in config else False
    handler = exp_strategy.csv_handler('{}.csv'.format(name.strip()))
    exp_strategy.minimize_latency(
        comp_path,
        comp_name,
        name,
        clock_period,
        part,
        standard,
        includes,
        defines,
        [handler],
        keep_env
    )

def handle_args(args):
    ini_file = Path(args.exp_file).resolve()
    if (not ini_file.exists()) or (not ini_file.is_file()):
        raise FileNotFoundError('File {} does not exists or is not a regular '
                                'file'.format(ini_file))
    config = configparser.ConfigParser()
    config.read(ini_file)
    logger = logging.getLogger('vrs_log')
    handler = logging.StreamHandler(stdout)
    log_level = logging.DEBUG if args.debug else logging.INFO
    logger.setLevel(log_level)
    handler.setLevel(log_level)
    logger.addHandler(handler)
    for s in config.sections():
        run_metaexp(s, config[s], ini_file.parent)


def main():
    parser = argparse.ArgumentParser(description="Perform synthesis and"
                                     "implementation of vivado HLS components")
    parser.add_argument("--debug", "-d", action="store_true", help="Activate "
                        "debug output")
    parser.add_argument("exp_file", help="Experiment description ini file")

    args = parser.parse_args()
    handle_args(args)


if __name__ == "__main__":
    main()
